from __future__ import print_function, division, absolute_import
from __future__ import unicode_literals
from fontTools.misc.testTools import getXML
from fontTools.otlLib import builder
from fontTools.ttLib.tables import otTables
from itertools import chain
import unittest


class BuilderTest(unittest.TestCase):
    GLYPHS = (".notdef space zero one two three four five six "
              "A B C a b c grave acute cedilla f_f_i f_i c_t").split()
    GLYPHMAP = {name: num for num, name in enumerate(GLYPHS)}

    ANCHOR1 = builder.buildAnchor(11, -11)
    ANCHOR2 = builder.buildAnchor(22, -22)
    ANCHOR3 = builder.buildAnchor(33, -33)

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None

    def test_buildAnchor_format1(self):
        anchor = builder.buildAnchor(23, 42)
        self.assertEqual(getXML(anchor.toXML),
                         ['<Anchor Format="1">',
                          '  <XCoordinate value="23"/>',
                          '  <YCoordinate value="42"/>',
                          '</Anchor>'])

    def test_buildAnchor_format2(self):
        anchor = builder.buildAnchor(23, 42, point=17)
        self.assertEqual(getXML(anchor.toXML),
                         ['<Anchor Format="2">',
                          '  <XCoordinate value="23"/>',
                          '  <YCoordinate value="42"/>',
                          '  <AnchorPoint value="17"/>',
                          '</Anchor>'])

    def test_buildAnchor_format3(self):
        anchor = builder.buildAnchor(
            23, 42,
            deviceX=builder.buildDevice({1: 1, 0: 0}),
            deviceY=builder.buildDevice({7: 7}))
        self.assertEqual(getXML(anchor.toXML),
                         ['<Anchor Format="3">',
                          '  <XCoordinate value="23"/>',
                          '  <YCoordinate value="42"/>',
                          '  <XDeviceTable>',
                          '    <StartSize value="0"/>',
                          '    <EndSize value="1"/>',
                          '    <DeltaFormat value="1"/>',
                          '    <DeltaValue value="[0, 1]"/>',
                          '  </XDeviceTable>',
                          '  <YDeviceTable>',
                          '    <StartSize value="7"/>',
                          '    <EndSize value="7"/>',
                          '    <DeltaFormat value="2"/>',
                          '    <DeltaValue value="[7]"/>',
                          '  </YDeviceTable>',
                          '</Anchor>'])

    def test_buildAttachList(self):
        attachList = builder.buildAttachList({
            "zero": [23, 7], "one": [1],
        }, self.GLYPHMAP)
        self.assertEqual(getXML(attachList.toXML),
                         ['<AttachList>',
                          '  <Coverage>',
                          '    <Glyph value="zero"/>',
                          '    <Glyph value="one"/>',
                          '  </Coverage>',
                          '  <!-- GlyphCount=2 -->',
                          '  <AttachPoint index="0">',
                          '    <!-- PointCount=2 -->',
                          '    <PointIndex index="0" value="7"/>',
                          '    <PointIndex index="1" value="23"/>',
                          '  </AttachPoint>',
                          '  <AttachPoint index="1">',
                          '    <!-- PointCount=1 -->',
                          '    <PointIndex index="0" value="1"/>',
                          '  </AttachPoint>',
                          '</AttachList>'])

    def test_buildAttachList_empty(self):
        self.assertIsNone(builder.buildAttachList({}, self.GLYPHMAP))

    def test_buildAttachPoint(self):
        attachPoint = builder.buildAttachPoint([7, 3])
        self.assertEqual(getXML(attachPoint.toXML),
                         ['<AttachPoint>',
                          '  <!-- PointCount=2 -->',
                          '  <PointIndex index="0" value="3"/>',
                          '  <PointIndex index="1" value="7"/>',
                          '</AttachPoint>'])

    def test_buildAttachPoint_empty(self):
        self.assertIsNone(builder.buildAttachPoint([]))

    def test_buildAttachPoint_duplicate(self):
        attachPoint = builder.buildAttachPoint([7, 3, 7])
        self.assertEqual(getXML(attachPoint.toXML),
                         ['<AttachPoint>',
                          '  <!-- PointCount=2 -->',
                          '  <PointIndex index="0" value="3"/>',
                          '  <PointIndex index="1" value="7"/>',
                          '</AttachPoint>'])


    def test_buildBaseArray(self):
        anchor = builder.buildAnchor
        baseArray = builder.buildBaseArray({
            "a": {2: anchor(300, 80)},
            "c": {1: anchor(300, 80), 2: anchor(300, -20)}
        }, numMarkClasses=4, glyphMap=self.GLYPHMAP)
        self.assertEqual(getXML(baseArray.toXML),
                         ['<BaseArray>',
                          '  <!-- BaseCount=2 -->',
                          '  <BaseRecord index="0">',
                          '    <BaseAnchor index="0" empty="1"/>',
                          '    <BaseAnchor index="1" empty="1"/>',
                          '    <BaseAnchor index="2" Format="1">',
                          '      <XCoordinate value="300"/>',
                          '      <YCoordinate value="80"/>',
                          '    </BaseAnchor>',
                          '    <BaseAnchor index="3" empty="1"/>',
                          '  </BaseRecord>',
                          '  <BaseRecord index="1">',
                          '    <BaseAnchor index="0" empty="1"/>',
                          '    <BaseAnchor index="1" Format="1">',
                          '      <XCoordinate value="300"/>',
                          '      <YCoordinate value="80"/>',
                          '    </BaseAnchor>',
                          '    <BaseAnchor index="2" Format="1">',
                          '      <XCoordinate value="300"/>',
                          '      <YCoordinate value="-20"/>',
                          '    </BaseAnchor>',
                          '    <BaseAnchor index="3" empty="1"/>',
                          '  </BaseRecord>',
                          '</BaseArray>'])

    def test_buildBaseRecord(self):
        a = builder.buildAnchor
        rec = builder.buildBaseRecord([a(500, -20), None, a(300, -15)])
        self.assertEqual(getXML(rec.toXML),
                         ['<BaseRecord>',
                          '  <BaseAnchor index="0" Format="1">',
                          '    <XCoordinate value="500"/>',
                          '    <YCoordinate value="-20"/>',
                          '  </BaseAnchor>',
                          '  <BaseAnchor index="1" empty="1"/>',
                          '  <BaseAnchor index="2" Format="1">',
                          '    <XCoordinate value="300"/>',
                          '    <YCoordinate value="-15"/>',
                          '  </BaseAnchor>',
                          '</BaseRecord>'])

    def test_buildCaretValueForCoord(self):
        caret = builder.buildCaretValueForCoord(500)
        self.assertEqual(getXML(caret.toXML),
                         ['<CaretValue Format="1">',
                          '  <Coordinate value="500"/>',
                          '</CaretValue>'])

    def test_buildCaretValueForPoint(self):
        caret = builder.buildCaretValueForPoint(23)
        self.assertEqual(getXML(caret.toXML),
                         ['<CaretValue Format="2">',
                          '  <CaretValuePoint value="23"/>',
                          '</CaretValue>'])

    def test_buildComponentRecord(self):
        a = builder.buildAnchor
        rec = builder.buildComponentRecord([a(500, -20), None, a(300, -15)])
        self.assertEqual(getXML(rec.toXML),
                         ['<ComponentRecord>',
                          '  <LigatureAnchor index="0" Format="1">',
                          '    <XCoordinate value="500"/>',
                          '    <YCoordinate value="-20"/>',
                          '  </LigatureAnchor>',
                          '  <LigatureAnchor index="1" empty="1"/>',
                          '  <LigatureAnchor index="2" Format="1">',
                          '    <XCoordinate value="300"/>',
                          '    <YCoordinate value="-15"/>',
                          '  </LigatureAnchor>',
                          '</ComponentRecord>'])

    def test_buildComponentRecord_empty(self):
        self.assertIsNone(builder.buildComponentRecord([]))

    def test_buildComponentRecord_None(self):
        self.assertIsNone(builder.buildComponentRecord(None))

    def test_buildCoverage(self):
        cov = builder.buildCoverage({"two", "four"}, {"two": 2, "four": 4})
        self.assertEqual(getXML(cov.toXML),
                         ['<Coverage>',
                          '  <Glyph value="two"/>',
                          '  <Glyph value="four"/>',
                          '</Coverage>'])

    def test_buildCursivePos(self):
        pos = builder.buildCursivePosSubtable({
            "two": (self.ANCHOR1, self.ANCHOR2),
            "four": (self.ANCHOR3, self.ANCHOR1)
        }, self.GLYPHMAP)
        self.assertEqual(getXML(pos.toXML),
                         ['<CursivePos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="two"/>',
                          '    <Glyph value="four"/>',
                          '  </Coverage>',
                          '  <!-- EntryExitCount=2 -->',
                          '  <EntryExitRecord index="0">',
                          '    <EntryAnchor Format="1">',
                          '      <XCoordinate value="11"/>',
                          '      <YCoordinate value="-11"/>',
                          '    </EntryAnchor>',
                          '    <ExitAnchor Format="1">',
                          '      <XCoordinate value="22"/>',
                          '      <YCoordinate value="-22"/>',
                          '    </ExitAnchor>',
                          '  </EntryExitRecord>',
                          '  <EntryExitRecord index="1">',
                          '    <EntryAnchor Format="1">',
                          '      <XCoordinate value="33"/>',
                          '      <YCoordinate value="-33"/>',
                          '    </EntryAnchor>',
                          '    <ExitAnchor Format="1">',
                          '      <XCoordinate value="11"/>',
                          '      <YCoordinate value="-11"/>',
                          '    </ExitAnchor>',
                          '  </EntryExitRecord>',
                          '</CursivePos>'])

    def test_buildDevice_format1(self):
        device = builder.buildDevice({1:1, 0:0})
        self.assertEqual(getXML(device.toXML),
                         ['<Device>',
                          '  <StartSize value="0"/>',
                          '  <EndSize value="1"/>',
                          '  <DeltaFormat value="1"/>',
                          '  <DeltaValue value="[0, 1]"/>',
                          '</Device>'])

    def test_buildDevice_format2(self):
        device = builder.buildDevice({2:2, 0:1, 1:0})
        self.assertEqual(getXML(device.toXML),
                         ['<Device>',
                          '  <StartSize value="0"/>',
                          '  <EndSize value="2"/>',
                          '  <DeltaFormat value="2"/>',
                          '  <DeltaValue value="[1, 0, 2]"/>',
                          '</Device>'])

    def test_buildDevice_format3(self):
        device = builder.buildDevice({5:3, 1:77})
        self.assertEqual(getXML(device.toXML),
                         ['<Device>',
                          '  <StartSize value="1"/>',
                          '  <EndSize value="5"/>',
                          '  <DeltaFormat value="3"/>',
                          '  <DeltaValue value="[77, 0, 0, 0, 3]"/>',
                          '</Device>'])

    def test_buildLigatureArray(self):
        anchor = builder.buildAnchor
        ligatureArray = builder.buildLigatureArray({
            "f_i": [{2: anchor(300, -20)}, {}],
            "c_t": [{}, {1: anchor(500, 350), 2: anchor(1300, -20)}]
        }, numMarkClasses=4, glyphMap=self.GLYPHMAP)
        self.assertEqual(getXML(ligatureArray.toXML),
                         ['<LigatureArray>',
                          '  <!-- LigatureCount=2 -->',
                          '  <LigatureAttach index="0">',  # f_i
                          '    <!-- ComponentCount=2 -->',
                          '    <ComponentRecord index="0">',
                          '      <LigatureAnchor index="0" empty="1"/>',
                          '      <LigatureAnchor index="1" empty="1"/>',
                          '      <LigatureAnchor index="2" Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="-20"/>',
                          '      </LigatureAnchor>',
                          '      <LigatureAnchor index="3" empty="1"/>',
                          '    </ComponentRecord>',
                          '    <ComponentRecord index="1">',
                          '      <LigatureAnchor index="0" empty="1"/>',
                          '      <LigatureAnchor index="1" empty="1"/>',
                          '      <LigatureAnchor index="2" empty="1"/>',
                          '      <LigatureAnchor index="3" empty="1"/>',
                          '    </ComponentRecord>',
                          '  </LigatureAttach>',
                          '  <LigatureAttach index="1">',
                          '    <!-- ComponentCount=2 -->',
                          '    <ComponentRecord index="0">',
                          '      <LigatureAnchor index="0" empty="1"/>',
                          '      <LigatureAnchor index="1" empty="1"/>',
                          '      <LigatureAnchor index="2" empty="1"/>',
                          '      <LigatureAnchor index="3" empty="1"/>',
                          '    </ComponentRecord>',
                          '    <ComponentRecord index="1">',
                          '      <LigatureAnchor index="0" empty="1"/>',
                          '      <LigatureAnchor index="1" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="350"/>',
                          '      </LigatureAnchor>',
                          '      <LigatureAnchor index="2" Format="1">',
                          '        <XCoordinate value="1300"/>',
                          '        <YCoordinate value="-20"/>',
                          '      </LigatureAnchor>',
                          '      <LigatureAnchor index="3" empty="1"/>',
                          '    </ComponentRecord>',
                          '  </LigatureAttach>',
                          '</LigatureArray>'])

    def test_buildLigatureAttach(self):
        anchor = builder.buildAnchor
        attach = builder.buildLigatureAttach([
            [anchor(500, -10), None],
            [None, anchor(300, -20), None]])
        self.assertEqual(getXML(attach.toXML),
                         ['<LigatureAttach>',
                          '  <!-- ComponentCount=2 -->',
                          '  <ComponentRecord index="0">',
                          '    <LigatureAnchor index="0" Format="1">',
                          '      <XCoordinate value="500"/>',
                          '      <YCoordinate value="-10"/>',
                          '    </LigatureAnchor>',
                          '    <LigatureAnchor index="1" empty="1"/>',
                          '  </ComponentRecord>',
                          '  <ComponentRecord index="1">',
                          '    <LigatureAnchor index="0" empty="1"/>',
                          '    <LigatureAnchor index="1" Format="1">',
                          '      <XCoordinate value="300"/>',
                          '      <YCoordinate value="-20"/>',
                          '    </LigatureAnchor>',
                          '    <LigatureAnchor index="2" empty="1"/>',
                          '  </ComponentRecord>',
                          '</LigatureAttach>'])

    def test_buildLigatureAttach_emptyComponents(self):
        attach = builder.buildLigatureAttach([[], None])
        self.assertEqual(getXML(attach.toXML),
                         ['<LigatureAttach>',
                          '  <!-- ComponentCount=2 -->',
                          '  <ComponentRecord index="0" empty="1"/>',
                          '  <ComponentRecord index="1" empty="1"/>',
                          '</LigatureAttach>'])

    def test_buildLigatureAttach_noComponents(self):
        attach = builder.buildLigatureAttach([])
        self.assertEqual(getXML(attach.toXML),
                         ['<LigatureAttach>',
                          '  <!-- ComponentCount=0 -->',
                          '</LigatureAttach>'])

    def test_buildLigCaretList(self):
        carets = builder.buildLigCaretList(
            {"f_f_i": [300, 600]}, {"c_t": [42]}, self.GLYPHMAP)
        self.assertEqual(getXML(carets.toXML),
                         ['<LigCaretList>',
                          '  <Coverage>',
                          '    <Glyph value="f_f_i"/>',
                          '    <Glyph value="c_t"/>',
                          '  </Coverage>',
                          '  <!-- LigGlyphCount=2 -->',
                          '  <LigGlyph index="0">',
                          '    <!-- CaretCount=2 -->',
                          '    <CaretValue index="0" Format="1">',
                          '      <Coordinate value="300"/>',
                          '    </CaretValue>',
                          '    <CaretValue index="1" Format="1">',
                          '      <Coordinate value="600"/>',
                          '    </CaretValue>',
                          '  </LigGlyph>',
                          '  <LigGlyph index="1">',
                          '    <!-- CaretCount=1 -->',
                          '    <CaretValue index="0" Format="2">',
                          '      <CaretValuePoint value="42"/>',
                          '    </CaretValue>',
                          '  </LigGlyph>',
                          '</LigCaretList>'])

    def test_buildLigCaretList_bothCoordsAndPointsForSameGlyph(self):
        carets = builder.buildLigCaretList(
            {"f_f_i": [300]}, {"f_f_i": [7]}, self.GLYPHMAP)
        self.assertEqual(getXML(carets.toXML),
                         ['<LigCaretList>',
                          '  <Coverage>',
                          '    <Glyph value="f_f_i"/>',
                          '  </Coverage>',
                          '  <!-- LigGlyphCount=1 -->',
                          '  <LigGlyph index="0">',
                          '    <!-- CaretCount=2 -->',
                          '    <CaretValue index="0" Format="1">',
                          '      <Coordinate value="300"/>',
                          '    </CaretValue>',
                          '    <CaretValue index="1" Format="2">',
                          '      <CaretValuePoint value="7"/>',
                          '    </CaretValue>',
                          '  </LigGlyph>',
                          '</LigCaretList>'])

    def test_buildLigCaretList_empty(self):
        self.assertIsNone(builder.buildLigCaretList({}, {}, self.GLYPHMAP))

    def test_buildLigCaretList_None(self):
        self.assertIsNone(builder.buildLigCaretList(None, None, self.GLYPHMAP))

    def test_buildLigGlyph_coords(self):
        lig = builder.buildLigGlyph([500, 800], None)
        self.assertEqual(getXML(lig.toXML),
                         ['<LigGlyph>',
                          '  <!-- CaretCount=2 -->',
                          '  <CaretValue index="0" Format="1">',
                          '    <Coordinate value="500"/>',
                          '  </CaretValue>',
                          '  <CaretValue index="1" Format="1">',
                          '    <Coordinate value="800"/>',
                          '  </CaretValue>',
                          '</LigGlyph>'])

    def test_buildLigGlyph_empty(self):
        self.assertIsNone(builder.buildLigGlyph([], []))

    def test_buildLigGlyph_None(self):
        self.assertIsNone(builder.buildLigGlyph(None, None))

    def test_buildLigGlyph_points(self):
        lig = builder.buildLigGlyph(None, [2])
        self.assertEqual(getXML(lig.toXML),
                         ['<LigGlyph>',
                          '  <!-- CaretCount=1 -->',
                          '  <CaretValue index="0" Format="2">',
                          '    <CaretValuePoint value="2"/>',
                          '  </CaretValue>',
                          '</LigGlyph>'])

    def test_buildLookup(self):
        s1 = builder.buildSingleSubstSubtable({"one": "two"})
        s2 = builder.buildSingleSubstSubtable({"three": "four"})
        lookup = builder.buildLookup([s1, s2], flags=7)
        self.assertEqual(getXML(lookup.toXML),
                         ['<Lookup>',
                          '  <LookupType value="1"/>',
                          '  <LookupFlag value="7"/>',
                          '  <!-- SubTableCount=2 -->',
                          '  <SingleSubst index="0">',
                          '    <Substitution in="one" out="two"/>',
                          '  </SingleSubst>',
                          '  <SingleSubst index="1">',
                          '    <Substitution in="three" out="four"/>',
                          '  </SingleSubst>',
                          '</Lookup>'])

    def test_buildLookup_badFlags(self):
        s = builder.buildSingleSubstSubtable({"one": "two"})
        self.assertRaisesRegex(
            AssertionError, "if markFilterSet is None, "
            "flags must not set LOOKUP_FLAG_USE_MARK_FILTERING_SET; "
            "flags=0x0010",
            builder.buildLookup, [s],
            builder.LOOKUP_FLAG_USE_MARK_FILTERING_SET, None)
        self.assertRaisesRegex(
            AssertionError, "if markFilterSet is not None, "
            "flags must set LOOKUP_FLAG_USE_MARK_FILTERING_SET; "
            "flags=0x0004",
            builder.buildLookup, [s],
            builder.LOOKUP_FLAG_IGNORE_LIGATURES, 777)

    def test_buildLookup_conflictingSubtableTypes(self):
        s1 = builder.buildSingleSubstSubtable({"one": "two"})
        s2 = builder.buildAlternateSubstSubtable({"one": ["two", "three"]})
        self.assertRaisesRegex(
            AssertionError, "all subtables must have the same LookupType",
            builder.buildLookup, [s1, s2])

    def test_buildLookup_noSubtables(self):
        self.assertIsNone(builder.buildLookup([]))
        self.assertIsNone(builder.buildLookup(None))
        self.assertIsNone(builder.buildLookup([None]))
        self.assertIsNone(builder.buildLookup([None, None]))

    def test_buildLookup_markFilterSet(self):
        s = builder.buildSingleSubstSubtable({"one": "two"})
        flags = (builder.LOOKUP_FLAG_RIGHT_TO_LEFT |
                 builder.LOOKUP_FLAG_USE_MARK_FILTERING_SET)
        lookup = builder.buildLookup([s], flags, markFilterSet=999)
        self.assertEqual(getXML(lookup.toXML),
                         ['<Lookup>',
                          '  <LookupType value="1"/>',
                          '  <LookupFlag value="17"/>',
                          '  <!-- SubTableCount=1 -->',
                          '  <SingleSubst index="0">',
                          '    <Substitution in="one" out="two"/>',
                          '  </SingleSubst>',
                          '  <MarkFilteringSet value="999"/>',
                          '</Lookup>'])

    def test_buildMarkArray(self):
        markArray = builder.buildMarkArray({
            "acute": (7, builder.buildAnchor(300, 800)),
            "grave": (2, builder.buildAnchor(10, 80))
        }, self.GLYPHMAP)
        self.assertLess(self.GLYPHMAP["grave"], self.GLYPHMAP["acute"])
        self.assertEqual(getXML(markArray.toXML),
                         ['<MarkArray>',
                          '  <!-- MarkCount=2 -->',
                          '  <MarkRecord index="0">',
                          '    <Class value="2"/>',
                          '    <MarkAnchor Format="1">',
                          '      <XCoordinate value="10"/>',
                          '      <YCoordinate value="80"/>',
                          '    </MarkAnchor>',
                          '  </MarkRecord>',
                          '  <MarkRecord index="1">',
                          '    <Class value="7"/>',
                          '    <MarkAnchor Format="1">',
                          '      <XCoordinate value="300"/>',
                          '      <YCoordinate value="800"/>',
                          '    </MarkAnchor>',
                          '  </MarkRecord>',
                          '</MarkArray>'])

    def test_buildMarkBasePosSubtable(self):
        anchor = builder.buildAnchor
        marks = {
            "acute": (0, anchor(300, 700)),
            "cedilla": (1, anchor(300, -100)),
            "grave": (0, anchor(300, 700))
        }
        bases = {
            # Make sure we can handle missing entries.
            "A": {},  # no entry for any markClass
            "B": {0: anchor(500, 900)},  # only markClass 0 specified
            "C": {1: anchor(500, -10)},  # only markClass 1 specified

            "a": {0: anchor(500, 400), 1: anchor(500, -20)},
            "b": {0: anchor(500, 800), 1: anchor(500, -20)}
        }
        table = builder.buildMarkBasePosSubtable(marks, bases, self.GLYPHMAP)
        self.assertEqual(getXML(table.toXML),
                         ['<MarkBasePos Format="1">',
                          '  <MarkCoverage>',
                          '    <Glyph value="grave"/>',
                          '    <Glyph value="acute"/>',
                          '    <Glyph value="cedilla"/>',
                          '  </MarkCoverage>',
                          '  <BaseCoverage>',
                          '    <Glyph value="A"/>',
                          '    <Glyph value="B"/>',
                          '    <Glyph value="C"/>',
                          '    <Glyph value="a"/>',
                          '    <Glyph value="b"/>',
                          '  </BaseCoverage>',
                          '  <!-- ClassCount=2 -->',
                          '  <MarkArray>',
                          '    <!-- MarkCount=3 -->',
                          '    <MarkRecord index="0">',  # grave
                          '      <Class value="0"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="700"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '    <MarkRecord index="1">',  # acute
                          '      <Class value="0"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="700"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '    <MarkRecord index="2">',  # cedilla
                          '      <Class value="1"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="-100"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '  </MarkArray>',
                          '  <BaseArray>',
                          '    <!-- BaseCount=5 -->',
                          '    <BaseRecord index="0">',  # A
                          '      <BaseAnchor index="0" empty="1"/>',
                          '      <BaseAnchor index="1" empty="1"/>',
                          '    </BaseRecord>',
                          '    <BaseRecord index="1">',  # B
                          '      <BaseAnchor index="0" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="900"/>',
                          '      </BaseAnchor>',
                          '      <BaseAnchor index="1" empty="1"/>',
                          '    </BaseRecord>',
                          '    <BaseRecord index="2">',  # C
                          '      <BaseAnchor index="0" empty="1"/>',
                          '      <BaseAnchor index="1" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="-10"/>',
                          '      </BaseAnchor>',
                          '    </BaseRecord>',
                          '    <BaseRecord index="3">',  # a
                          '      <BaseAnchor index="0" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="400"/>',
                          '      </BaseAnchor>',
                          '      <BaseAnchor index="1" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="-20"/>',
                          '      </BaseAnchor>',
                          '    </BaseRecord>',
                          '    <BaseRecord index="4">',  # b
                          '      <BaseAnchor index="0" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="800"/>',
                          '      </BaseAnchor>',
                          '      <BaseAnchor index="1" Format="1">',
                          '        <XCoordinate value="500"/>',
                          '        <YCoordinate value="-20"/>',
                          '      </BaseAnchor>',
                          '    </BaseRecord>',
                          '  </BaseArray>',
                          '</MarkBasePos>'])

    def test_buildMarkGlyphSetsDef(self):
        marksets = builder.buildMarkGlyphSetsDef(
            [{"acute", "grave"}, {"cedilla", "grave"}], self.GLYPHMAP)
        self.assertEqual(getXML(marksets.toXML),
                         ['<MarkGlyphSetsDef>',
                          '  <MarkSetTableFormat value="1"/>',
                          '  <!-- MarkSetCount=2 -->',
                          '  <Coverage index="0">',
                          '    <Glyph value="grave"/>',
                          '    <Glyph value="acute"/>',
                          '  </Coverage>',
                          '  <Coverage index="1">',
                          '    <Glyph value="grave"/>',
                          '    <Glyph value="cedilla"/>',
                          '  </Coverage>',
                          '</MarkGlyphSetsDef>'])

    def test_buildMarkGlyphSetsDef_empty(self):
        self.assertIsNone(builder.buildMarkGlyphSetsDef([], self.GLYPHMAP))

    def test_buildMarkGlyphSetsDef_None(self):
        self.assertIsNone(builder.buildMarkGlyphSetsDef(None, self.GLYPHMAP))

    def test_buildMarkLigPosSubtable(self):
        anchor = builder.buildAnchor
        marks = {
            "acute": (0, anchor(300, 700)),
            "cedilla": (1, anchor(300, -100)),
            "grave": (0, anchor(300, 700))
        }
        bases = {
            "f_i": [{}, {0: anchor(200, 400)}],  # nothing on f; only 1 on i
            "c_t": [
                {0: anchor(500, 600), 1: anchor(500, -20)},   # c
                {0: anchor(1300, 800), 1: anchor(1300, -20)}  # t
            ]
        }
        table = builder.buildMarkLigPosSubtable(marks, bases, self.GLYPHMAP)
        self.assertEqual(getXML(table.toXML),
                         ['<MarkLigPos Format="1">',
                          '  <MarkCoverage>',
                          '    <Glyph value="grave"/>',
                          '    <Glyph value="acute"/>',
                          '    <Glyph value="cedilla"/>',
                          '  </MarkCoverage>',
                          '  <LigatureCoverage>',
                          '    <Glyph value="f_i"/>',
                          '    <Glyph value="c_t"/>',
                          '  </LigatureCoverage>',
                          '  <!-- ClassCount=2 -->',
                          '  <MarkArray>',
                          '    <!-- MarkCount=3 -->',
                          '    <MarkRecord index="0">',
                          '      <Class value="0"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="700"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '    <MarkRecord index="1">',
                          '      <Class value="0"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="700"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '    <MarkRecord index="2">',
                          '      <Class value="1"/>',
                          '      <MarkAnchor Format="1">',
                          '        <XCoordinate value="300"/>',
                          '        <YCoordinate value="-100"/>',
                          '      </MarkAnchor>',
                          '    </MarkRecord>',
                          '  </MarkArray>',
                          '  <LigatureArray>',
                          '    <!-- LigatureCount=2 -->',
                          '    <LigatureAttach index="0">',
                          '      <!-- ComponentCount=2 -->',
                          '      <ComponentRecord index="0">',
                          '        <LigatureAnchor index="0" empty="1"/>',
                          '        <LigatureAnchor index="1" empty="1"/>',
                          '      </ComponentRecord>',
                          '      <ComponentRecord index="1">',
                          '        <LigatureAnchor index="0" Format="1">',
                          '          <XCoordinate value="200"/>',
                          '          <YCoordinate value="400"/>',
                          '        </LigatureAnchor>',
                          '        <LigatureAnchor index="1" empty="1"/>',
                          '      </ComponentRecord>',
                          '    </LigatureAttach>',
                          '    <LigatureAttach index="1">',
                          '      <!-- ComponentCount=2 -->',
                          '      <ComponentRecord index="0">',
                          '        <LigatureAnchor index="0" Format="1">',
                          '          <XCoordinate value="500"/>',
                          '          <YCoordinate value="600"/>',
                          '        </LigatureAnchor>',
                          '        <LigatureAnchor index="1" Format="1">',
                          '          <XCoordinate value="500"/>',
                          '          <YCoordinate value="-20"/>',
                          '        </LigatureAnchor>',
                          '      </ComponentRecord>',
                          '      <ComponentRecord index="1">',
                          '        <LigatureAnchor index="0" Format="1">',
                          '          <XCoordinate value="1300"/>',
                          '          <YCoordinate value="800"/>',
                          '        </LigatureAnchor>',
                          '        <LigatureAnchor index="1" Format="1">',
                          '          <XCoordinate value="1300"/>',
                          '          <YCoordinate value="-20"/>',
                          '        </LigatureAnchor>',
                          '      </ComponentRecord>',
                          '    </LigatureAttach>',
                          '  </LigatureArray>',
                          '</MarkLigPos>'])

    def test_buildMarkRecord(self):
        rec = builder.buildMarkRecord(17, builder.buildAnchor(500, -20))
        self.assertEqual(getXML(rec.toXML),
                         ['<MarkRecord>',
                          '  <Class value="17"/>',
                          '  <MarkAnchor Format="1">',
                          '    <XCoordinate value="500"/>',
                          '    <YCoordinate value="-20"/>',
                          '  </MarkAnchor>',
                          '</MarkRecord>'])

    def test_buildMark2Record(self):
        a = builder.buildAnchor
        rec = builder.buildMark2Record([a(500, -20), None, a(300, -15)])
        self.assertEqual(getXML(rec.toXML),
                         ['<Mark2Record>',
                          '  <Mark2Anchor index="0" Format="1">',
                          '    <XCoordinate value="500"/>',
                          '    <YCoordinate value="-20"/>',
                          '  </Mark2Anchor>',
                          '  <Mark2Anchor index="1" empty="1"/>',
                          '  <Mark2Anchor index="2" Format="1">',
                          '    <XCoordinate value="300"/>',
                          '    <YCoordinate value="-15"/>',
                          '  </Mark2Anchor>',
                          '</Mark2Record>'])

    def test_buildPairPosClassesSubtable(self):
        d20 = builder.buildValue({"XPlacement": -20})
        d50 = builder.buildValue({"XPlacement": -50})
        d0 = builder.buildValue({})
        d8020 = builder.buildValue({"XPlacement": -80, "YPlacement": -20})
        subtable = builder.buildPairPosClassesSubtable({
            (tuple("A",), tuple(["zero"])): (d0, d50),
            (tuple("A",), tuple(["one", "two"])):  (None, d20),
            (tuple(["B", "C"]), tuple(["zero"])): (d8020, d50),
        }, self.GLYPHMAP)
        self.assertEqual(getXML(subtable.toXML),
                         ['<PairPos Format="2">',
                          '  <Coverage>',
                          '    <Glyph value="A"/>',
                          '    <Glyph value="B"/>',
                          '    <Glyph value="C"/>',
                          '  </Coverage>',
                          '  <ValueFormat1 value="3"/>',
                          '  <ValueFormat2 value="1"/>',
                          '  <ClassDef1>',
                          '    <ClassDef glyph="A" class="1"/>',
                          '  </ClassDef1>',
                          '  <ClassDef2>',
                          '    <ClassDef glyph="one" class="1"/>',
                          '    <ClassDef glyph="two" class="1"/>',
                          '    <ClassDef glyph="zero" class="2"/>',
                          '  </ClassDef2>',
                          '  <!-- Class1Count=2 -->',
                          '  <!-- Class2Count=3 -->',
                          '  <Class1Record index="0">',
                          '    <Class2Record index="0">',
                          '    </Class2Record>',
                          '    <Class2Record index="1">',
                          '    </Class2Record>',
                          '    <Class2Record index="2">',
                          '      <Value1 XPlacement="-80" YPlacement="-20"/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </Class2Record>',
                          '  </Class1Record>',
                          '  <Class1Record index="1">',
                          '    <Class2Record index="0">',
                          '    </Class2Record>',
                          '    <Class2Record index="1">',
                          '      <Value2 XPlacement="-20"/>',
                          '    </Class2Record>',
                          '    <Class2Record index="2">',
                          '      <Value1/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </Class2Record>',
                          '  </Class1Record>',
                          '</PairPos>'])

    def test_buildPairPosGlyphs(self):
        d50 = builder.buildValue({"XPlacement": -50})
        d8020 = builder.buildValue({"XPlacement": -80, "YPlacement": -20})
        subtables = builder.buildPairPosGlyphs({
            ("A", "zero"): (None, d50),
            ("A", "one"):  (d8020, d50),
        }, self.GLYPHMAP)
        self.assertEqual(sum([getXML(t.toXML) for t in subtables], []),
                         ['<PairPos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="A"/>',
                          '  </Coverage>',
                          '  <ValueFormat1 value="0"/>',
                          '  <ValueFormat2 value="1"/>',
                          '  <!-- PairSetCount=1 -->',
                          '  <PairSet index="0">',
                          '    <!-- PairValueCount=1 -->',
                          '    <PairValueRecord index="0">',
                          '      <SecondGlyph value="zero"/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </PairValueRecord>',
                          '  </PairSet>',
                          '</PairPos>',
                          '<PairPos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="A"/>',
                          '  </Coverage>',
                          '  <ValueFormat1 value="3"/>',
                          '  <ValueFormat2 value="1"/>',
                          '  <!-- PairSetCount=1 -->',
                          '  <PairSet index="0">',
                          '    <!-- PairValueCount=1 -->',
                          '    <PairValueRecord index="0">',
                          '      <SecondGlyph value="one"/>',
                          '      <Value1 XPlacement="-80" YPlacement="-20"/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </PairValueRecord>',
                          '  </PairSet>',
                          '</PairPos>'])

    def test_buildPairPosGlyphsSubtable(self):
        d20 = builder.buildValue({"XPlacement": -20})
        d50 = builder.buildValue({"XPlacement": -50})
        d0 = builder.buildValue({})
        d8020 = builder.buildValue({"XPlacement": -80, "YPlacement": -20})
        subtable = builder.buildPairPosGlyphsSubtable({
            ("A", "zero"): (d0, d50),
            ("A", "one"):  (None, d20),
            ("B", "five"): (d8020, d50),
        }, self.GLYPHMAP)
        self.assertEqual(getXML(subtable.toXML),
                         ['<PairPos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="A"/>',
                          '    <Glyph value="B"/>',
                          '  </Coverage>',
                          '  <ValueFormat1 value="3"/>',
                          '  <ValueFormat2 value="1"/>',
                          '  <!-- PairSetCount=2 -->',
                          '  <PairSet index="0">',
                          '    <!-- PairValueCount=2 -->',
                          '    <PairValueRecord index="0">',
                          '      <SecondGlyph value="zero"/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </PairValueRecord>',
                          '    <PairValueRecord index="1">',
                          '      <SecondGlyph value="one"/>',
                          '      <Value2 XPlacement="-20"/>',
                          '    </PairValueRecord>',
                          '  </PairSet>',
                          '  <PairSet index="1">',
                          '    <!-- PairValueCount=1 -->',
                          '    <PairValueRecord index="0">',
                          '      <SecondGlyph value="five"/>',
                          '      <Value1 XPlacement="-80" YPlacement="-20"/>',
                          '      <Value2 XPlacement="-50"/>',
                          '    </PairValueRecord>',
                          '  </PairSet>',
                          '</PairPos>'])

    def test_buildSinglePos(self):
        subtables = builder.buildSinglePos({
            "one": builder.buildValue({"XPlacement": 500}),
            "two": builder.buildValue({"XPlacement": 500}),
            "three": builder.buildValue({"XPlacement": 200}),
            "four": builder.buildValue({"XPlacement": 400}),
            "five": builder.buildValue({"XPlacement": 500}),
            "six": builder.buildValue({"YPlacement": -6}),
        }, self.GLYPHMAP)
        self.assertEqual(sum([getXML(t.toXML) for t in subtables], []),
                         ['<SinglePos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="one"/>',
                          '    <Glyph value="two"/>',
                          '    <Glyph value="five"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="1"/>',
                          '  <Value XPlacement="500"/>',
                          '</SinglePos>',
                          '<SinglePos Format="2">',
                          '  <Coverage>',
                          '    <Glyph value="three"/>',
                          '    <Glyph value="four"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="1"/>',
                          '  <!-- ValueCount=2 -->',
                          '  <Value index="0" XPlacement="200"/>',
                          '  <Value index="1" XPlacement="400"/>',
                          '</SinglePos>',
                          '<SinglePos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="six"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="2"/>',
                          '  <Value YPlacement="-6"/>',
                          '</SinglePos>'])

    def test_buildSinglePos_ValueFormat0(self):
        subtables = builder.buildSinglePos({
            "zero": builder.buildValue({})
        }, self.GLYPHMAP)
        self.assertEqual(sum([getXML(t.toXML) for t in subtables], []),
                         ['<SinglePos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="zero"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="0"/>',
                          '</SinglePos>'])

    def test_buildSinglePosSubtable_format1(self):
        subtable = builder.buildSinglePosSubtable({
            "one": builder.buildValue({"XPlacement": 777}),
            "two": builder.buildValue({"XPlacement": 777}),
        }, self.GLYPHMAP)
        self.assertEqual(getXML(subtable.toXML),
                         ['<SinglePos Format="1">',
                          '  <Coverage>',
                          '    <Glyph value="one"/>',
                          '    <Glyph value="two"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="1"/>',
                          '  <Value XPlacement="777"/>',
                          '</SinglePos>'])

    def test_buildSinglePosSubtable_format2(self):
        subtable = builder.buildSinglePosSubtable({
            "one": builder.buildValue({"XPlacement": 777}),
            "two": builder.buildValue({"YPlacement": -888}),
        }, self.GLYPHMAP)
        self.assertEqual(getXML(subtable.toXML),
                         ['<SinglePos Format="2">',
                          '  <Coverage>',
                          '    <Glyph value="one"/>',
                          '    <Glyph value="two"/>',
                          '  </Coverage>',
                          '  <ValueFormat value="3"/>',
                          '  <!-- ValueCount=2 -->',
                          '  <Value index="0" XPlacement="777"/>',
                          '  <Value index="1" YPlacement="-888"/>',
                          '</SinglePos>'])

    def test_buildValue(self):
        value = builder.buildValue({"XPlacement": 7, "YPlacement": 23})
        func = lambda writer, font: value.toXML(writer, font, valueName="Val")
        self.assertEqual(getXML(func),
                         ['<Val XPlacement="7" YPlacement="23"/>'])

    def test_getLigatureKey(self):
        components = lambda s: [tuple(word) for word in s.split()]
        c = components("fi fl ff ffi fff")
        c.sort(key=builder._getLigatureKey)
        self.assertEqual(c, components("fff ffi ff fi fl"))

    def test_getSinglePosValueKey(self):
        device = builder.buildDevice({10:1, 11:3})
        a1 = builder.buildValue({"XPlacement": 500, "XPlaDevice": device})
        a2 = builder.buildValue({"XPlacement": 500, "XPlaDevice": device})
        b = builder.buildValue({"XPlacement": 500})
        keyA1 = builder._getSinglePosValueKey(a1)
        keyA2 = builder._getSinglePosValueKey(a1)
        keyB = builder._getSinglePosValueKey(b)
        self.assertEqual(keyA1, keyA2)
        self.assertEqual(hash(keyA1), hash(keyA2))
        self.assertNotEqual(keyA1, keyB)
        self.assertNotEqual(hash(keyA1), hash(keyB))


class ClassDefBuilderTest(unittest.TestCase):
    def test_build_usingClass0(self):
        b = builder.ClassDefBuilder(useClass0=True)
        b.add({"aa", "bb"})
        b.add({"a", "b"})
        b.add({"c"})
        b.add({"e", "f", "g", "h"})
        cdef = b.build()
        self.assertIsInstance(cdef, otTables.ClassDef)
        self.assertEqual(cdef.classDefs, {
            "a": 2,
            "b": 2,
            "c": 3,
            "aa": 1,
            "bb": 1
        })

    def test_build_notUsingClass0(self):
        b = builder.ClassDefBuilder(useClass0=False)
        b.add({"a", "b"})
        b.add({"c"})
        b.add({"e", "f", "g", "h"})
        cdef = b.build()
        self.assertIsInstance(cdef, otTables.ClassDef)
        self.assertEqual(cdef.classDefs, {
            "a": 2,
            "b": 2,
            "c": 3,
            "e": 1,
            "f": 1,
            "g": 1,
            "h": 1
        })

    def test_canAdd(self):
        b = builder.ClassDefBuilder(useClass0=True)
        b.add({"a", "b", "c", "d"})
        b.add({"e", "f"})
        self.assertTrue(b.canAdd({"a", "b", "c", "d"}))
        self.assertTrue(b.canAdd({"e", "f"}))
        self.assertTrue(b.canAdd({"g", "h", "i"}))
        self.assertFalse(b.canAdd({"b", "c", "d"}))
        self.assertFalse(b.canAdd({"a", "b", "c", "d", "e", "f"}))
        self.assertFalse(b.canAdd({"d", "e", "f"}))
        self.assertFalse(b.canAdd({"f"}))


if __name__ == "__main__":
    import sys
    sys.exit(unittest.main())



from __future__ import print_function, division, absolute_import, unicode_literals
from fontTools.misc.py23 import *
from fontTools import ttLib
import unittest
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable, table__c_m_a_p

class CmapSubtableTest(unittest.TestCase):

	def makeSubtable(self, cmapFormat, platformID, platEncID, langID):
		subtable = CmapSubtable.newSubtable(cmapFormat)
		subtable.platformID, subtable.platEncID, subtable.language = (platformID, platEncID, langID)
		return subtable

	def test_toUnicode_utf16be(self):
		subtable = self.makeSubtable(4, 0, 2, 7)
		self.assertEqual("utf_16_be", subtable.getEncoding())
		self.assertEqual(True, subtable.isUnicode())

	def test_toUnicode_macroman(self):
		subtable = self.makeSubtable(4, 1, 0, 7)  # MacRoman
		self.assertEqual("mac_roman", subtable.getEncoding())
		self.assertEqual(False, subtable.isUnicode())

	def test_toUnicode_macromanian(self):
		subtable = self.makeSubtable(4, 1, 0, 37)  # Mac Romanian
		self.assertNotEqual(None, subtable.getEncoding())
		self.assertEqual(False, subtable.isUnicode())

	def test_extended_mac_encodings(self):
		subtable = self.makeSubtable(4, 1, 1, 0) # Mac Japanese
		self.assertNotEqual(None, subtable.getEncoding())
		self.assertEqual(False, subtable.isUnicode())

	def test_extended_unknown(self):
		subtable = self.makeSubtable(4, 10, 11, 12)
		self.assertEqual(subtable.getEncoding(), None)
		self.assertEqual(subtable.getEncoding("ascii"), "ascii")
		self.assertEqual(subtable.getEncoding(default="xyz"), "xyz")

	def test_decompile_4(self):
		subtable = CmapSubtable.newSubtable(4)
		font = ttLib.TTFont()
		font.setGlyphOrder([])
		subtable.decompile(b'\0' * 3 + b'\x10' + b'\0' * 12, font)

	def test_decompile_12(self):
		subtable = CmapSubtable.newSubtable(12)
		font = ttLib.TTFont()
		font.setGlyphOrder([])
		subtable.decompile(b'\0' * 7 + b'\x10' + b'\0' * 8, font)

	def test_buildReversed(self):
		c4 = self.makeSubtable(4, 3, 1, 0)
		c4.cmap = {0x0041:'A', 0x0391:'A'}
		c12 = self.makeSubtable(12, 3, 10, 0)
		c12.cmap = {0x10314: 'u10314'}
		cmap = table__c_m_a_p()
		cmap.tables = [c4, c12]
		self.assertEqual(cmap.buildReversed(), {'A':{0x0041, 0x0391}, 'u10314':{0x10314}})

	def test_getBestCmap(self):
		c4 = self.makeSubtable(4, 3, 1, 0)
		c4.cmap = {0x0041:'A', 0x0391:'A'}
		c12 = self.makeSubtable(12, 3, 10, 0)
		c12.cmap = {0x10314: 'u10314'}
		cmap = table__c_m_a_p()
		cmap.tables = [c4, c12]
		self.assertEqual(cmap.getBestCmap(), {0x10314: 'u10314'})
		self.assertEqual(cmap.getBestCmap(cmapPreferences=[(3, 1)]), {0x0041:'A', 0x0391:'A'})
		self.assertEqual(cmap.getBestCmap(cmapPreferences=[(0, 4)]), None)

	def test_font_getBestCmap(self):
		c4 = self.makeSubtable(4, 3, 1, 0)
		c4.cmap = {0x0041:'A', 0x0391:'A'}
		c12 = self.makeSubtable(12, 3, 10, 0)
		c12.cmap = {0x10314: 'u10314'}
		cmap = table__c_m_a_p()
		cmap.tables = [c4, c12]
		font = ttLib.TTFont()
		font["cmap"] = cmap
		self.assertEqual(font.getBestCmap(), {0x10314: 'u10314'})
		self.assertEqual(font.getBestCmap(cmapPreferences=[(3, 1)]), {0x0041:'A', 0x0391:'A'})
		self.assertEqual(font.getBestCmap(cmapPreferences=[(0, 4)]), None)


if __name__ == "__main__":
	import sys
	sys.exit(unittest.main())



# SVG Path specification parser.
# This is an adaptation from 'svg.path' by Lennart Regebro (@regebro),
# modified so that the parser takes a FontTools Pen object instead of
# returning a list of svg.path Path objects.
# The original code can be found at:
# https://github.com/regebro/svg.path/blob/4f9b6e3/src/svg/path/parser.py
# Copyright (c) 2013-2014 Lennart Regebro
# License: MIT

from __future__ import (
    print_function, division, absolute_import, unicode_literals)
from fontTools.misc.py23 import *
import re

COMMANDS = set('MmZzLlHhVvCcSsQqTtAa')
UPPERCASE = set('MZLHVCSQTA')

COMMAND_RE = re.compile("([MmZzLlHhVvCcSsQqTtAa])")
FLOAT_RE = re.compile("[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?")


def _tokenize_path(pathdef):
    for x in COMMAND_RE.split(pathdef):
        if x in COMMANDS:
            yield x
        for token in FLOAT_RE.findall(x):
            yield token


def parse_path(pathdef, pen, current_pos=(0, 0)):
    """ Parse SVG path definition (i.e. "d" attribute of <path> elements)
    and call a 'pen' object's moveTo, lineTo, curveTo, qCurveTo and closePath
    methods.

    If 'current_pos' (2-float tuple) is provided, the initial moveTo will
    be relative to that instead being absolute.

    Arc segments (commands "A" or "a") are not currently supported, and raise
    NotImplementedError.
    """
    # In the SVG specs, initial movetos are absolute, even if
    # specified as 'm'. This is the default behavior here as well.
    # But if you pass in a current_pos variable, the initial moveto
    # will be relative to that current_pos. This is useful.
    current_pos = complex(*current_pos)

    elements = list(_tokenize_path(pathdef))
    # Reverse for easy use of .pop()
    elements.reverse()

    start_pos = None
    command = None
    last_control = None

    while elements:

        if elements[-1] in COMMANDS:
            # New command.
            last_command = command  # Used by S and T
            command = elements.pop()
            absolute = command in UPPERCASE
            command = command.upper()
        else:
            # If this element starts with numbers, it is an implicit command
            # and we don't change the command. Check that it's allowed:
            if command is None:
                raise ValueError("Unallowed implicit command in %s, position %s" % (
                    pathdef, len(pathdef.split()) - len(elements)))
            last_command = command  # Used by S and T

        if command == 'M':
            # Moveto command.
            x = elements.pop()
            y = elements.pop()
            pos = float(x) + float(y) * 1j
            if absolute:
                current_pos = pos
            else:
                current_pos += pos

            # M is not preceded by Z; it's an open subpath
            if start_pos is not None:
                pen.endPath()

            pen.moveTo((current_pos.real, current_pos.imag))

            # when M is called, reset start_pos
            # This behavior of Z is defined in svg spec:
            # http://www.w3.org/TR/SVG/paths.html#PathDataClosePathCommand
            start_pos = current_pos

            # Implicit moveto commands are treated as lineto commands.
            # So we set command to lineto here, in case there are
            # further implicit commands after this moveto.
            command = 'L'

        elif command == 'Z':
            # Close path
            if current_pos != start_pos:
                pen.lineTo((start_pos.real, start_pos.imag))
            pen.closePath()
            current_pos = start_pos
            start_pos = None
            command = None  # You can't have implicit commands after closing.

        elif command == 'L':
            x = elements.pop()
            y = elements.pop()
            pos = float(x) + float(y) * 1j
            if not absolute:
                pos += current_pos
            pen.lineTo((pos.real, pos.imag))
            current_pos = pos

        elif command == 'H':
            x = elements.pop()
            pos = float(x) + current_pos.imag * 1j
            if not absolute:
                pos += current_pos.real
            pen.lineTo((pos.real, pos.imag))
            current_pos = pos

        elif command == 'V':
            y = elements.pop()
            pos = current_pos.real + float(y) * 1j
            if not absolute:
                pos += current_pos.imag * 1j
            pen.lineTo((pos.real, pos.imag))
            current_pos = pos

        elif command == 'C':
            control1 = float(elements.pop()) + float(elements.pop()) * 1j
            control2 = float(elements.pop()) + float(elements.pop()) * 1j
            end = float(elements.pop()) + float(elements.pop()) * 1j

            if not absolute:
                control1 += current_pos
                control2 += current_pos
                end += current_pos

            pen.curveTo((control1.real, control1.imag),
                        (control2.real, control2.imag),
                        (end.real, end.imag))
            current_pos = end
            last_control = control2

        elif command == 'S':
            # Smooth curve. First control point is the "reflection" of
            # the second control point in the previous path.

            if last_command not in 'CS':
                # If there is no previous command or if the previous command
                # was not an C, c, S or s, assume the first control point is
                # coincident with the current point.
                control1 = current_pos
            else:
                # The first control point is assumed to be the reflection of
                # the second control point on the previous command relative
                # to the current point.
                control1 = current_pos + current_pos - last_control

            control2 = float(elements.pop()) + float(elements.pop()) * 1j
            end = float(elements.pop()) + float(elements.pop()) * 1j

            if not absolute:
                control2 += current_pos
                end += current_pos

            pen.curveTo((control1.real, control1.imag),
                        (control2.real, control2.imag),
                        (end.real, end.imag))
            current_pos = end
            last_control = control2

        elif command == 'Q':
            control = float(elements.pop()) + float(elements.pop()) * 1j
            end = float(elements.pop()) + float(elements.pop()) * 1j

            if not absolute:
                control += current_pos
                end += current_pos

            pen.qCurveTo((control.real, control.imag), (end.real, end.imag))
            current_pos = end
            last_control = control

        elif command == 'T':
            # Smooth curve. Control point is the "reflection" of
            # the second control point in the previous path.

            if last_command not in 'QT':
                # If there is no previous command or if the previous command
                # was not an Q, q, T or t, assume the first control point is
                # coincident with the current point.
                control = current_pos
            else:
                # The control point is assumed to be the reflection of
                # the control point on the previous command relative
                # to the current point.
                control = current_pos + current_pos - last_control

            end = float(elements.pop()) + float(elements.pop()) * 1j

            if not absolute:
                end += current_pos

            pen.qCurveTo((control.real, control.imag), (end.real, end.imag))
            current_pos = end
            last_control = control

        elif command == 'A':
            raise NotImplementedError('arcs are not supported')

    # no final Z command, it's an open path
    if start_pos is not None:
        pen.endPath()



from __future__ import print_function, division, absolute_import
from fontTools.misc.py23 import *
from fontTools.misc.fixedTools import fixedToFloat, floatToFixed
from fontTools.misc.textTools import safeEval
import array
import io
import logging
import struct
import sys


# https://www.microsoft.com/typography/otspec/otvarcommonformats.htm

EMBEDDED_PEAK_TUPLE = 0x8000
INTERMEDIATE_REGION = 0x4000
PRIVATE_POINT_NUMBERS = 0x2000

DELTAS_ARE_ZERO = 0x80
DELTAS_ARE_WORDS = 0x40
DELTA_RUN_COUNT_MASK = 0x3f

POINTS_ARE_WORDS = 0x80
POINT_RUN_COUNT_MASK = 0x7f

TUPLES_SHARE_POINT_NUMBERS = 0x8000
TUPLE_COUNT_MASK = 0x0fff
TUPLE_INDEX_MASK = 0x0fff

log = logging.getLogger(__name__)


class TupleVariation(object):
	def __init__(self, axes, coordinates):
		self.axes = axes.copy()
		self.coordinates = coordinates[:]

	def __repr__(self):
		axes = ",".join(sorted(["%s=%s" % (name, value) for (name, value) in self.axes.items()]))
		return "<TupleVariation %s %s>" % (axes, self.coordinates)

	def __eq__(self, other):
		return self.coordinates == other.coordinates and self.axes == other.axes

	def getUsedPoints(self):
		result = set()
		for i, point in enumerate(self.coordinates):
			if point is not None:
				result.add(i)
		return result

	def hasImpact(self):
		"""Returns True if this TupleVariation has any visible impact.

		If the result is False, the TupleVariation can be omitted from the font
		without making any visible difference.
		"""
		for c in self.coordinates:
			if c is not None:
				return True
		return False

	def toXML(self, writer, axisTags):
		writer.begintag("tuple")
		writer.newline()
		for axis in axisTags:
			value = self.axes.get(axis)
			if value is not None:
				minValue, value, maxValue = (float(v) for v in value)
				defaultMinValue = min(value, 0.0)  # -0.3 --> -0.3; 0.7 --> 0.0
				defaultMaxValue = max(value, 0.0)  # -0.3 -->  0.0; 0.7 --> 0.7
				if minValue == defaultMinValue and maxValue == defaultMaxValue:
					writer.simpletag("coord", axis=axis, value=value)
				else:
					writer.simpletag("coord", axis=axis, value=value, min=minValue, max=maxValue)
				writer.newline()
		wrote_any_deltas = False
		for i, delta in enumerate(self.coordinates):
			if type(delta) == tuple and len(delta) == 2:
				writer.simpletag("delta", pt=i, x=delta[0], y=delta[1])
				writer.newline()
				wrote_any_deltas = True
			elif type(delta) == int:
				writer.simpletag("delta", cvt=i, value=delta)
				writer.newline()
				wrote_any_deltas = True
			elif delta is not None:
				log.error("bad delta format")
				writer.comment("bad delta #%d" % i)
				writer.newline()
				wrote_any_deltas = True
		if not wrote_any_deltas:
			writer.comment("no deltas")
			writer.newline()
		writer.endtag("tuple")
		writer.newline()

	def fromXML(self, name, attrs, _content):
		if name == "coord":
			axis = attrs["axis"]
			value = float(attrs["value"])
			defaultMinValue = min(value, 0.0)  # -0.3 --> -0.3; 0.7 --> 0.0
			defaultMaxValue = max(value, 0.0)  # -0.3 -->  0.0; 0.7 --> 0.7
			minValue = float(attrs.get("min", defaultMinValue))
			maxValue = float(attrs.get("max", defaultMaxValue))
			self.axes[axis] = (minValue, value, maxValue)
		elif name == "delta":
			if "pt" in attrs:
				point = safeEval(attrs["pt"])
				x = safeEval(attrs["x"])
				y = safeEval(attrs["y"])
				self.coordinates[point] = (x, y)
			elif "cvt" in attrs:
				cvt = safeEval(attrs["cvt"])
				value = safeEval(attrs["value"])
				self.coordinates[cvt] = value
			else:
				log.warning("bad delta format: %s" %
				            ", ".join(sorted(attrs.keys())))

	def compile(self, axisTags, sharedCoordIndices, sharedPoints):
		tupleData = []

		assert all(tag in axisTags for tag in self.axes.keys()), ("Unknown axis tag found.", self.axes.keys(), axisTags)

		coord = self.compileCoord(axisTags)
		if coord in sharedCoordIndices:
			flags = sharedCoordIndices[coord]
		else:
			flags = EMBEDDED_PEAK_TUPLE
			tupleData.append(coord)

		intermediateCoord = self.compileIntermediateCoord(axisTags)
		if intermediateCoord is not None:
			flags |= INTERMEDIATE_REGION
			tupleData.append(intermediateCoord)

		points = self.getUsedPoints()
		if sharedPoints == points:
			# Only use the shared points if they are identical to the actually used points
			auxData = self.compileDeltas(sharedPoints)
			usesSharedPoints = True
		else:
			flags |= PRIVATE_POINT_NUMBERS
			numPointsInGlyph = len(self.coordinates)
			auxData = self.compilePoints(points, numPointsInGlyph) + self.compileDeltas(points)
			usesSharedPoints = False

		tupleData = struct.pack('>HH', len(auxData), flags) + bytesjoin(tupleData)
		return (tupleData, auxData, usesSharedPoints)

	def compileCoord(self, axisTags):
		result = []
		for axis in axisTags:
			_minValue, value, _maxValue = self.axes.get(axis, (0.0, 0.0, 0.0))
			result.append(struct.pack(">h", floatToFixed(value, 14)))
		return bytesjoin(result)

	def compileIntermediateCoord(self, axisTags):
		needed = False
		for axis in axisTags:
			minValue, value, maxValue = self.axes.get(axis, (0.0, 0.0, 0.0))
			defaultMinValue = min(value, 0.0)  # -0.3 --> -0.3; 0.7 --> 0.0
			defaultMaxValue = max(value, 0.0)  # -0.3 -->  0.0; 0.7 --> 0.7
			if (minValue != defaultMinValue) or (maxValue != defaultMaxValue):
				needed = True
				break
		if not needed:
			return None
		minCoords = []
		maxCoords = []
		for axis in axisTags:
			minValue, value, maxValue = self.axes.get(axis, (0.0, 0.0, 0.0))
			minCoords.append(struct.pack(">h", floatToFixed(minValue, 14)))
			maxCoords.append(struct.pack(">h", floatToFixed(maxValue, 14)))
		return bytesjoin(minCoords + maxCoords)

	@staticmethod
	def decompileCoord_(axisTags, data, offset):
		coord = {}
		pos = offset
		for axis in axisTags:
			coord[axis] = fixedToFloat(struct.unpack(">h", data[pos:pos+2])[0], 14)
			pos += 2
		return coord, pos

	@staticmethod
	def compilePoints(points, numPointsInGlyph):
		# If the set consists of all points in the glyph, it gets encoded with
		# a special encoding: a single zero byte.
		if len(points) == numPointsInGlyph:
			return b"\0"

		# In the 'gvar' table, the packing of point numbers is a little surprising.
		# It consists of multiple runs, each being a delta-encoded list of integers.
		# For example, the point set {17, 18, 19, 20, 21, 22, 23} gets encoded as
		# [6, 17, 1, 1, 1, 1, 1, 1]. The first value (6) is the run length minus 1.
		# There are two types of runs, with values being either 8 or 16 bit unsigned
		# integers.
		points = list(points)
		points.sort()
		numPoints = len(points)

		# The binary representation starts with the total number of points in the set,
		# encoded into one or two bytes depending on the value.
		if numPoints < 0x80:
			result = [bytechr(numPoints)]
		else:
			result = [bytechr((numPoints >> 8) | 0x80) + bytechr(numPoints & 0xff)]

		MAX_RUN_LENGTH = 127
		pos = 0
		lastValue = 0
		while pos < numPoints:
			run = io.BytesIO()
			runLength = 0
			useByteEncoding = None
			while pos < numPoints and runLength <= MAX_RUN_LENGTH:
				curValue = points[pos]
				delta = curValue - lastValue
				if useByteEncoding is None:
					useByteEncoding = 0 <= delta <= 0xff
				if useByteEncoding and (delta > 0xff or delta < 0):
					# we need to start a new run (which will not use byte encoding)
					break
				# TODO This never switches back to a byte-encoding from a short-encoding.
				# That's suboptimal.
				if useByteEncoding:
					run.write(bytechr(delta))
				else:
					run.write(bytechr(delta >> 8))
					run.write(bytechr(delta & 0xff))
				lastValue = curValue
				pos += 1
				runLength += 1
			if useByteEncoding:
				runHeader = bytechr(runLength - 1)
			else:
				runHeader = bytechr((runLength - 1) | POINTS_ARE_WORDS)
			result.append(runHeader)
			result.append(run.getvalue())

		return bytesjoin(result)

	@staticmethod
	def decompilePoints_(numPoints, data, offset, tableTag):
		"""(numPoints, data, offset, tableTag) --> ([point1, point2, ...], newOffset)"""
		assert tableTag in ('cvar', 'gvar')
		pos = offset
		numPointsInData = byteord(data[pos])
		pos += 1
		if (numPointsInData & POINTS_ARE_WORDS) != 0:
			numPointsInData = (numPointsInData & POINT_RUN_COUNT_MASK) << 8 | byteord(data[pos])
			pos += 1
		if numPointsInData == 0:
			return (range(numPoints), pos)

		result = []
		while len(result) < numPointsInData:
			runHeader = byteord(data[pos])
			pos += 1
			numPointsInRun = (runHeader & POINT_RUN_COUNT_MASK) + 1
			point = 0
			if (runHeader & POINTS_ARE_WORDS) != 0:
				points = array.array("H")
				pointsSize = numPointsInRun * 2
			else:
				points = array.array("B")
				pointsSize = numPointsInRun
			points.fromstring(data[pos:pos+pointsSize])
			if sys.byteorder != "big":
				points.byteswap()

			assert len(points) == numPointsInRun
			pos += pointsSize

			result.extend(points)

		# Convert relative to absolute
		absolute = []
		current = 0
		for delta in result:
			current += delta
			absolute.append(current)
		result = absolute
		del absolute

		badPoints = {str(p) for p in result if p < 0 or p >= numPoints}
		if badPoints:
			log.warning("point %s out of range in '%s' table" %
			            (",".join(sorted(badPoints)), tableTag))
		return (result, pos)

	def compileDeltas(self, points):
		deltaX = []
		deltaY = []
		for p in sorted(list(points)):
			c = self.coordinates[p]
			if type(c) is tuple and len(c) == 2:
				deltaX.append(c[0])
				deltaY.append(c[1])
			elif type(c) is int:
				deltaX.append(c)
			elif c is not None:
				raise ValueError("invalid type of delta: %s" % type(c))
		return self.compileDeltaValues_(deltaX) + self.compileDeltaValues_(deltaY)

	@staticmethod
	def compileDeltaValues_(deltas):
		"""[value1, value2, value3, ...] --> bytestring

		Emits a sequence of runs. Each run starts with a
		byte-sized header whose 6 least significant bits
		(header & 0x3F) indicate how many values are encoded
		in this run. The stored length is the actual length
		minus one; run lengths are thus in the range [1..64].
		If the header byte has its most significant bit (0x80)
		set, all values in this run are zero, and no data
		follows. Otherwise, the header byte is followed by
		((header & 0x3F) + 1) signed values.  If (header &
		0x40) is clear, the delta values are stored as signed
		bytes; if (header & 0x40) is set, the delta values are
		signed 16-bit integers.
		"""  # Explaining the format because the 'gvar' spec is hard to understand.
		stream = io.BytesIO()
		pos = 0
		while pos < len(deltas):
			value = deltas[pos]
			if value == 0:
				pos = TupleVariation.encodeDeltaRunAsZeroes_(deltas, pos, stream)
			elif value >= -128 and value <= 127:
				pos = TupleVariation.encodeDeltaRunAsBytes_(deltas, pos, stream)
			else:
				pos = TupleVariation.encodeDeltaRunAsWords_(deltas, pos, stream)
		return stream.getvalue()

	@staticmethod
	def encodeDeltaRunAsZeroes_(deltas, offset, stream):
		runLength = 0
		pos = offset
		numDeltas = len(deltas)
		while pos < numDeltas and runLength < 64 and deltas[pos] == 0:
			pos += 1
			runLength += 1
		assert runLength >= 1 and runLength <= 64
		stream.write(bytechr(DELTAS_ARE_ZERO | (runLength - 1)))
		return pos

	@staticmethod
	def encodeDeltaRunAsBytes_(deltas, offset, stream):
		runLength = 0
		pos = offset
		numDeltas = len(deltas)
		while pos < numDeltas and runLength < 64:
			value = deltas[pos]
			if value < -128 or value > 127:
				break
			# Within a byte-encoded run of deltas, a single zero
			# is best stored literally as 0x00 value. However,
			# if are two or more zeroes in a sequence, it is
			# better to start a new run. For example, the sequence
			# of deltas [15, 15, 0, 15, 15] becomes 6 bytes
			# (04 0F 0F 00 0F 0F) when storing the zero value
			# literally, but 7 bytes (01 0F 0F 80 01 0F 0F)
			# when starting a new run.
			if value == 0 and pos+1 < numDeltas and deltas[pos+1] == 0:
				break
			pos += 1
			runLength += 1
		assert runLength >= 1 and runLength <= 64
		stream.write(bytechr(runLength - 1))
		for i in range(offset, pos):
			stream.write(struct.pack('b', round(deltas[i])))
		return pos

	@staticmethod
	def encodeDeltaRunAsWords_(deltas, offset, stream):
		runLength = 0
		pos = offset
		numDeltas = len(deltas)
		while pos < numDeltas and runLength < 64:
			value = deltas[pos]
			# Within a word-encoded run of deltas, it is easiest
			# to start a new run (with a different encoding)
			# whenever we encounter a zero value. For example,
			# the sequence [0x6666, 0, 0x7777] needs 7 bytes when
			# storing the zero literally (42 66 66 00 00 77 77),
			# and equally 7 bytes when starting a new run
			# (40 66 66 80 40 77 77).
			if value == 0:
				break

			# Within a word-encoded run of deltas, a single value
			# in the range (-128..127) should be encoded literally
			# because it is more compact. For example, the sequence
			# [0x6666, 2, 0x7777] becomes 7 bytes when storing
			# the value literally (42 66 66 00 02 77 77), but 8 bytes
			# when starting a new run (40 66 66 00 02 40 77 77).
			isByteEncodable = lambda value: value >= -128 and value <= 127
			if isByteEncodable(value) and pos+1 < numDeltas and isByteEncodable(deltas[pos+1]):
				break
			pos += 1
			runLength += 1
		assert runLength >= 1 and runLength <= 64
		stream.write(bytechr(DELTAS_ARE_WORDS | (runLength - 1)))
		for i in range(offset, pos):
			stream.write(struct.pack('>h', round(deltas[i])))
		return pos

	@staticmethod
	def decompileDeltas_(numDeltas, data, offset):
		"""(numDeltas, data, offset) --> ([delta, delta, ...], newOffset)"""
		result = []
		pos = offset
		while len(result) < numDeltas:
			runHeader = byteord(data[pos])
			pos += 1
			numDeltasInRun = (runHeader & DELTA_RUN_COUNT_MASK) + 1
			if (runHeader & DELTAS_ARE_ZERO) != 0:
				result.extend([0] * numDeltasInRun)
			else:
				if (runHeader & DELTAS_ARE_WORDS) != 0:
					deltas = array.array("h")
					deltasSize = numDeltasInRun * 2
				else:
					deltas = array.array("b")
					deltasSize = numDeltasInRun
				deltas.fromstring(data[pos:pos+deltasSize])
				if sys.byteorder != "big":
					deltas.byteswap()
				assert len(deltas) == numDeltasInRun
				pos += deltasSize
				result.extend(deltas)
		assert len(result) == numDeltas
		return (result, pos)

	@staticmethod
	def getTupleSize_(flags, axisCount):
		size = 4
		if (flags & EMBEDDED_PEAK_TUPLE) != 0:
			size += axisCount * 2
		if (flags & INTERMEDIATE_REGION) != 0:
			size += axisCount * 4
		return size


def decompileSharedTuples(axisTags, sharedTupleCount, data, offset):
	result = []
	for _ in range(sharedTupleCount):
		t, offset = TupleVariation.decompileCoord_(axisTags, data, offset)
		result.append(t)
	return result


def compileSharedTuples(axisTags, variations):
	coordCount = {}
	for var in variations:
		coord = var.compileCoord(axisTags)
		coordCount[coord] = coordCount.get(coord, 0) + 1
	sharedCoords = [(count, coord)
					for (coord, count) in coordCount.items() if count > 1]
	sharedCoords.sort(reverse=True)
	MAX_NUM_SHARED_COORDS = TUPLE_INDEX_MASK + 1
	sharedCoords = sharedCoords[:MAX_NUM_SHARED_COORDS]
	return [c[1] for c in sharedCoords]  # Strip off counts.


def compileTupleVariationStore(variations, pointCount,
                               axisTags, sharedTupleIndices,
                               useSharedPoints=True):
	variations = [v for v in variations if v.hasImpact()]
	if len(variations) == 0:
		return (0, b"", b"")

	# Each glyph variation tuples modifies a set of control points. To
	# indicate which exact points are getting modified, a single tuple
	# can either refer to a shared set of points, or the tuple can
	# supply its private point numbers.  Because the impact of sharing
	# can be positive (no need for a private point list) or negative
	# (need to supply 0,0 deltas for unused points), it is not obvious
	# how to determine which tuples should take their points from the
	# shared pool versus have their own. Perhaps we should resort to
	# brute force, and try all combinations? However, if a glyph has n
	# variation tuples, we would need to try 2^n combinations (because
	# each tuple may or may not be part of the shared set). How many
	# variations tuples do glyphs have?
	#
	#   Skia.ttf: {3: 1, 5: 11, 6: 41, 7: 62, 8: 387, 13: 1, 14: 3}
	#   JamRegular.ttf: {3: 13, 4: 122, 5: 1, 7: 4, 8: 1, 9: 1, 10: 1}
	#   BuffaloGalRegular.ttf: {1: 16, 2: 13, 4: 2, 5: 4, 6: 19, 7: 1, 8: 3, 9: 8}
	#   (Reading example: In Skia.ttf, 41 glyphs have 6 variation tuples).
	#

	# Is this even worth optimizing? If we never use a shared point
	# list, the private lists will consume 112K for Skia, 5K for
	# BuffaloGalRegular, and 15K for JamRegular. If we always use a
	# shared point list, the shared lists will consume 16K for Skia,
	# 3K for BuffaloGalRegular, and 10K for JamRegular. However, in
	# the latter case the delta arrays will become larger, but I
	# haven't yet measured by how much. From gut feeling (which may be
	# wrong), the optimum is to share some but not all points;
	# however, then we would need to try all combinations.
	#
	# For the time being, we try two variants and then pick the better one:
	# (a) each tuple supplies its own private set of points;
	# (b) all tuples refer to a shared set of points, which consists of
	#     "every control point in the glyph that has explicit deltas".
	usedPoints = set()
	for v in variations:
		usedPoints |= v.getUsedPoints()
	tuples = []
	data = []
	someTuplesSharePoints = False
	sharedPointVariation = None # To keep track of a variation that uses shared points
	for v in variations:
		privateTuple, privateData, _ = v.compile(
			axisTags, sharedTupleIndices, sharedPoints=None)
		sharedTuple, sharedData, usesSharedPoints = v.compile(
			axisTags, sharedTupleIndices, sharedPoints=usedPoints)
		if useSharedPoints and (len(sharedTuple) + len(sharedData)) < (len(privateTuple) + len(privateData)):
			tuples.append(sharedTuple)
			data.append(sharedData)
			someTuplesSharePoints |= usesSharedPoints
			sharedPointVariation = v
		else:
			tuples.append(privateTuple)
			data.append(privateData)
	if someTuplesSharePoints:
		# Use the last of the variations that share points for compiling the packed point data
		data = sharedPointVariation.compilePoints(usedPoints, len(sharedPointVariation.coordinates)) + bytesjoin(data)
		tupleVariationCount = TUPLES_SHARE_POINT_NUMBERS | len(tuples)
	else:
		data = bytesjoin(data)
		tupleVariationCount = len(tuples)
	tuples = bytesjoin(tuples)
	return tupleVariationCount, tuples, data


def decompileTupleVariationStore(tableTag, axisTags,
                                 tupleVariationCount, pointCount, sharedTuples,
							     data, pos, dataPos):
	numAxes = len(axisTags)
	result = []
	if (tupleVariationCount & TUPLES_SHARE_POINT_NUMBERS) != 0:
		sharedPoints, dataPos = TupleVariation.decompilePoints_(
			pointCount, data, dataPos, tableTag)
	else:
		sharedPoints = []
	for _ in range(tupleVariationCount & TUPLE_COUNT_MASK):
		dataSize, flags = struct.unpack(">HH", data[pos:pos+4])
		tupleSize = TupleVariation.getTupleSize_(flags, numAxes)
		tupleData = data[pos : pos + tupleSize]
		pointDeltaData = data[dataPos : dataPos + dataSize]
		result.append(decompileTupleVariation_(
			pointCount, sharedTuples, sharedPoints,
			tableTag, axisTags, tupleData, pointDeltaData))
		pos += tupleSize
		dataPos += dataSize
	return result


def decompileTupleVariation_(pointCount, sharedTuples, sharedPoints,
							 tableTag, axisTags, data, tupleData):
	assert tableTag in ("cvar", "gvar"), tableTag
	flags = struct.unpack(">H", data[2:4])[0]
	pos = 4
	if (flags & EMBEDDED_PEAK_TUPLE) == 0:
		peak = sharedTuples[flags & TUPLE_INDEX_MASK]
	else:
		peak, pos = TupleVariation.decompileCoord_(axisTags, data, pos)
	if (flags & INTERMEDIATE_REGION) != 0:
		start, pos = TupleVariation.decompileCoord_(axisTags, data, pos)
		end, pos = TupleVariation.decompileCoord_(axisTags, data, pos)
	else:
		start, end = inferRegion_(peak)
	axes = {}
	for axis in axisTags:
		region = start[axis], peak[axis], end[axis]
		if region != (0.0, 0.0, 0.0):
			axes[axis] = region
	pos = 0
	if (flags & PRIVATE_POINT_NUMBERS) != 0:
		points, pos = TupleVariation.decompilePoints_(
			pointCount, tupleData, pos, tableTag)
	else:
		points = sharedPoints

	deltas = [None] * pointCount

	if tableTag == "cvar":
		deltas_cvt, pos = TupleVariation.decompileDeltas_(
			len(points), tupleData, pos)
		for p, delta in zip(points, deltas_cvt):
			if 0 <= p < pointCount:
				deltas[p] = delta

	elif tableTag == "gvar":
		deltas_x, pos = TupleVariation.decompileDeltas_(
			len(points), tupleData, pos)
		deltas_y, pos = TupleVariation.decompileDeltas_(
			len(points), tupleData, pos)
		for p, x, y in zip(points, deltas_x, deltas_y):
			if 0 <= p < pointCount:
				deltas[p] = (x, y)

	return TupleVariation(axes, deltas)


def inferRegion_(peak):
	"""Infer start and end for a (non-intermediate) region

	This helper function computes the applicability region for
	variation tuples whose INTERMEDIATE_REGION flag is not set in the
	TupleVariationHeader structure.  Variation tuples apply only to
	certain regions of the variation space; outside that region, the
	tuple has no effect.  To make the binary encoding more compact,
	TupleVariationHeaders can omit the intermediateStartTuple and
	intermediateEndTuple fields.
    """
	start, end = {}, {}
	for (axis, value) in peak.items():
		start[axis] = min(value, 0.0)  # -0.3 --> -0.3; 0.7 --> 0.0
		end[axis] = max(value, 0.0)  # -0.3 -->  0.0; 0.7 --> 0.7
	return (start, end)



from __future__ import print_function, division, absolute_import
from __future__ import unicode_literals
from fontTools.varLib import designspace
import os
import unittest


class DesignspaceTest(unittest.TestCase):
    def test_load(self):
        self.maxDiff = None
        self.assertEqual(
            designspace.load(_getpath("Designspace.designspace")),

                {'sources':
                  [{'location': {'weight': 0.0},
                    'groups': {'copy': True},
                    'filename': 'DesignspaceTest-Light.ufo',
                    'info': {'copy': True},
                    'name': 'master_1',
                    'lib': {'copy': True}},
                   {'location': {'weight': 1.0},
                    'name': 'master_2',
                    'filename': 'DesignspaceTest-Bold.ufo'}],

                 'instances':
                  [{'location': {'weight': 0.5},
                    'familyname': 'DesignspaceTest',
                    'filename': 'instance/DesignspaceTest-Medium.ufo',
                    'kerning': {},
                    'info': {},
                    'stylename': 'Medium'}],

                 'axes':
                  [{'name': 'weight',
                    'map': [{'input': 0.0, 'output': 10.0},
                            {'input': 401.0, 'output': 66.0},
                            {'input': 1000.0, 'output': 990.0}],
                    'tag': 'wght',
                    'maximum': 1000.0,
                    'minimum': 0.0,
                    'default': 0.0},
                   {'maximum': 1000.0,
                    'default': 250.0,
                    'minimum': 0.0,
                    'name': 'width',
                    'tag': 'wdth'},
                   {'name': 'contrast',
                    'tag': 'cntr',
                    'maximum': 100.0,
                    'minimum': 0.0,
                    'default': 0.0,
                    'labelname': {'de': 'Kontrast', 'en': 'Contrast'}}]
                }
        )

    def test_load2(self):
        self.assertEqual(
            designspace.load(_getpath("Designspace2.designspace")),
                    {'sources': [], 'instances': [{}]})


def _getpath(testfile):
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", testfile)


if __name__ == "__main__":
    import sys
    sys.exit(unittest.main())



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod

"""GUI font inspector.
"""

from __future__ import print_function, division, absolute_import
from fontTools.misc.py23 import *
from fontTools import misc, ttLib, cffLib
try:
    from gi import pygtkcompat
except ImportError:
    pygtkcompat = None

if pygtkcompat is not None:
    pygtkcompat.enable()
    pygtkcompat.enable_gtk(version='3.0')
import gtk
import sys


class Row(object):
	def __init__(self, parent, index, key, value, font):
		self._parent = parent
		self._index = index
		self._key = key
		self._value = value
		self._font = font

		if isinstance(value, ttLib.TTFont):
			self._add_font(value)
			return

		if not isinstance(value, basestring):
			# Try sequences
			is_sequence = True
			try:
				len(value)
				iter(value)
				# It's hard to differentiate list-type sequences
				# from dict-type ones.  Try fetching item 0.
				value[0]
			except (TypeError, AttributeError, KeyError, IndexError):
				is_sequence = False
			if is_sequence:
				self._add_list(key, value)
				return
		if hasattr(value, '__dict__'):
			self._add_object(key, value)
			return
		if hasattr(value, 'items'):
			self._add_dict(key, value)
			return

		if isinstance(value, basestring):
			self._value_str = '"'+value+'"'
			self._children = []
			return

		# Everything else
		self._children = []

	def _filter_items(self):
		items = []
		for k,v in self._items:
			if isinstance(v, ttLib.TTFont):
				continue
			if k in ['reader', 'file', 'tableTag', 'compileStatus', 'recurse']:
				continue
			if isinstance(k, basestring) and k[0] == '_':
				continue
			items.append((k,v))
		self._items = items

	def _add_font(self, font):
		self._items = [(tag,font[tag]) for tag in font.keys()]

	def _add_object(self, key, value):
		# Make sure item is decompiled
		try:
			value.asdf # Any better way?!
		except (AttributeError, KeyError, TypeError, ttLib.TTLibError):
			pass
		if isinstance(value, ttLib.getTableModule('glyf').Glyph):
			# Glyph type needs explicit expanding to be useful
			value.expand(self._font['glyf'])
		if isinstance(value, misc.psCharStrings.T2CharString):
			try:
				value.decompile()
			except TypeError:  # Subroutines can't be decompiled
				pass
		if isinstance(value, cffLib.BaseDict):
			for k in value.rawDict.keys():
				getattr(value, k)
		if isinstance(value, cffLib.Index):
			# Load all items
			for i in range(len(value)):
				value[i]
			# Discard offsets as should not be needed anymore
			if hasattr(value, 'offsets'):
				del value.offsets

		self._value_str = value.__class__.__name__
		if isinstance(value, ttLib.tables.DefaultTable.DefaultTable):
			self._value_str += ' (%d Bytes)' % self._font.reader.tables[key].length
		self._items = sorted(value.__dict__.items())
		self._filter_items()

	def _add_dict(self, key, value):
		self._value_str = '%s of %d items' % (value.__class__.__name__, len(value))
		self._items = sorted(value.items())

	def _add_list(self, key, value):
		if len(value) and len(value) <= 32:
			self._value_str = str(value)
		else:
			self._value_str = '%s of %d items' % (value.__class__.__name__, len(value))
		self._items = list(enumerate(value))

	def __len__(self):
		if hasattr(self, '_children'):
			return len(self._children)
		if hasattr(self, '_items'):
			return len(self._items)
		assert False

	def _ensure_children(self):
		if hasattr(self, '_children'):
			return
		children = []
		for i,(k,v) in enumerate(self._items):
			children.append(Row(self, i, k, v, self._font))
		self._children = children
		del self._items

	def __getitem__(self, n):
		if n >= len(self):
			return None
		if not hasattr(self, '_children'):
			self._children = [None] * len(self)
		c = self._children[n]
		if c is None:
			k,v = self._items[n]
			c = self._children[n] = Row(self, n, k, v, self._font)
			self._items[n] = None
		return c

	def get_parent(self):
		return self._parent

	def get_index(self):
		return self._index

	def get_key(self):
		return self._key

	def get_value(self):
		return self._value

	def get_value_str(self):
		if hasattr(self,'_value_str'):
			return self._value_str
		return str(self._value)

class FontTreeModel(gtk.GenericTreeModel):

	__gtype_name__ = 'FontTreeModel'

	def __init__(self, font):
		super(FontTreeModel, self).__init__()
		self._columns = (str, str)
		self.font = font
		self._root = Row(None, 0, "font", font, font)

	def on_get_flags(self):
		return 0

	def on_get_n_columns(self):
		return len(self._columns)

	def on_get_column_type(self, index):
		return self._columns[index]

	def on_get_iter(self, path):
		rowref = self._root
		while path:
			rowref = rowref[path[0]]
			path = path[1:]
		return rowref

	def on_get_path(self, rowref):
		path = []
		while rowref != self._root:
			path.append(rowref.get_index())
			rowref = rowref.get_parent()
		path.reverse()
		return tuple(path)

	def on_get_value(self, rowref, column):
		if column == 0:
			return rowref.get_key()
		else:
			return rowref.get_value_str()

	def on_iter_next(self, rowref):
		return rowref.get_parent()[rowref.get_index() + 1]

	def on_iter_children(self, rowref):
		return rowref[0]

	def on_iter_has_child(self, rowref):
		return bool(len(rowref))

	def on_iter_n_children(self, rowref):
		return len(rowref)

	def on_iter_nth_child(self, rowref, n):
		if not rowref: rowref = self._root
		return rowref[n]

	def on_iter_parent(self, rowref):
		return rowref.get_parent()

class Inspect(object):

	def _delete_event(self, widget, event, data=None):
		gtk.main_quit()
		return False

	def __init__(self, fontfile):

		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title("%s - pyftinspect" % fontfile)
		self.window.connect("delete_event", self._delete_event)
		self.window.set_size_request(400, 600)

		self.scrolled_window = gtk.ScrolledWindow()
		self.window.add(self.scrolled_window)

		self.font = ttLib.TTFont(fontfile, lazy=True)
		self.treemodel = FontTreeModel(self.font)
		self.treeview = gtk.TreeView(self.treemodel)
		#self.treeview.set_reorderable(True)

		for i in range(2):
			col_name = ('Key', 'Value')[i]
			col = gtk.TreeViewColumn(col_name)
			col.set_sort_column_id(-1)
			self.treeview.append_column(col)

			cell = gtk.CellRendererText()
			col.pack_start(cell, True)
			col.add_attribute(cell, 'text', i)

		self.treeview.set_search_column(1)
		self.scrolled_window.add(self.treeview)
		self.window.show_all()

def main(args=None):
	if args is None:
		args = sys.argv[1:]
	if len(args) < 1:
		print("usage: pyftinspect font...", file=sys.stderr)
		return 1
	for arg in args:
		Inspect(arg)
	gtk.main()

if __name__ == "__main__":
	sys.exit(main())
